from __future__ import annotations

import concurrent
import logging
import re
from concurrent.futures import ThreadPoolExecutor, as_completed
from functools import partial
from time import sleep

import requests
from deep_translator import GoogleTranslator
from deep_translator.exceptions import RequestError, TooManyRequests
from langdetect import detect
from tqdm import tqdm

from plagdef.model.models import Document

WEBSHARE_PROXIES = "https://proxy.webshare.io/proxy/list/download/rzeoaimkyxecclzdabargzhwodnrgicaedlyppgc/-/http" \
                   "/username/direct/"

log = logging.getLogger(__name__)


def detect_lang(docs: set[Document]) -> None:
    for doc in docs:
        doc.lang = detect(doc.text) if doc.text else None


def docs_in_other_langs(docs: set[Document], expected_lang: str) -> set[Document]:
    return {doc for doc in docs if doc.lang != expected_lang}


def translate(docs: set[Document], target_lang: str) -> set[Document]:
    proxies = _get_proxies()
    translated = set()
    for doc in tqdm(docs, desc='Translating', unit='doc'):
        if doc.lang != target_lang:
            if len(doc.text) < 50000:
                _translate(doc, target_lang, proxies)
                translated.add(doc)
            else:
                log.warning(f'Skipping translation of {doc} because its text length is greater than 50k chars.')
    return translated


def _translate(doc: Document, target_lang: str, proxies: list[str]) -> None:
    # The limit of Google Translate Web is less than 5000 chars per request
    chunks = {(i, chunk) for i, chunk in enumerate(_split_text_at_punct(doc.text, 4999))}
    translated = set()
    while len(chunks) > len(translated):
        chunks = chunks.difference(translated)
        executor = ThreadPoolExecutor(len(proxies))
        futures = [executor.submit(partial(_translate_chunk, target_lang=target_lang,
                                           proxy=proxies[i % len(proxies)]), chunk) for i, chunk in enumerate(chunks)]
        try:
            for future in as_completed(futures, timeout=30):
                result = future.result()
                if result:
                    translated.add(result)
        except concurrent.futures.TimeoutError:
            pass
        finally:
            executor.shutdown(wait=False, cancel_futures=True)
    chunk_texts = [chunk[1] for chunk in sorted(translated, key=lambda chunk: chunk[0])]
    doc.text = "".join(chunk_texts)
    doc.lang = target_lang


def _split_text_at_punct(text: str, max_len: int, chunks: list[str] = None) -> list[str]:
    if chunks is None:
        chunks = []
    if len(text) <= max_len:
        return [*chunks, text]
    match = re.search(r'(?s:.*)[\.!\?]\s', text[:max_len])
    chunk = text[:match.end()] if match else text[:max_len]
    chunks.append(chunk)
    return _split_text_at_punct(text[match.end() if match else max_len:], max_len, chunks)


def _translate_chunk(chunk: tuple[int, str], target_lang: str, proxy: str):
    try:
        sleep(1)
        return chunk[0], GoogleTranslator(target=target_lang, proxies={"https": proxy}).translate(text=chunk[1])
    except (RequestError, TooManyRequests):
        return None


def _get_proxies() -> list[str]:
    proxies = []
    for proxy in requests.get(WEBSHARE_PROXIES).text.splitlines():
        address = proxy.split(":")
        proxies.append(f"http://{address[2]}:{address[3]}@{address[0]}:{address[1]}")
    return proxies



from __future__ import annotations

import os
import signal
import sys
from pathlib import Path

import click
from click import UsageError

from plagdef import services
from plagdef.config import settings
from plagdef.model.models import DocumentPairMatches, Document
from plagdef.model.reporting import generate_text_report
from plagdef.repositories import DocumentFileRepository, DocumentPairMatchesJsonRepository, DocumentPairRepository


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(prog_name='PlagDef', package_name='plagdef')
@click.argument('docdir', type=(click.Path(exists=True), bool))
@click.option('lang', '--lang', '-l', required=True, help='Bibliographical language code of the documents\' language.')
@click.option('archive_docdir', '--archive-docs', '-a', type=(click.Path(exists=True), bool),
              help='Directory containing documents which should not be compared among themselves but only with '
                   'the documents in <DOCDIR>.')
@click.option('common_docdir', '--common-docs', '-c', type=(click.Path(exists=True), bool),
              help='Directory containing documents with sentences which should be excluded from detection.')
@click.option('download_path', '--download-path', '-d', type=(click.Path(exists=True)),
              help='If set, it enables the download of external sources into the defined folder.')
@click.option('sim_th', '--similarity-threshold', '-s', type=click.FloatRange(0, 1), default=0.6,
              help='Similarity threshold for text matching, defaults to 0.6. Lower values may increase sensitivity, '
                   'higher ones can improve precision.')
@click.option('ocr', '--ocr', '-o', type=bool, default=True, help='Use OCR for PDFs with poor text layers.'
                                                                  'May improve text extraction but significantly '
                                                                  'reduces performance.')
@click.option('jsondir', '--json', '-j', type=click.Path(), help='Output directory for JSON reports.')
def cli(docdir: tuple[click.Path, bool], lang: str, ocr: bool, common_docdir: [click.Path, bool],
        archive_docdir: [click.Path, bool], sim_th: float, jsondir: click.Path, download_path: click.Path):
    """
    \b
    PlagDef supports plagiarism detection for student assignments.
    It must be provided a <DOCDIR> <RECURSIVE> tuple and the language option.
    The chosen directory must contain at least two documents.
    For instance if you would like to recursively search <DOCDIR> the correct command looks like this:
    `plagdef <DOCDIR> True`
    """
    settings.update({'lang': lang, 'ocr': ocr, 'min_cos_sim': sim_th, 'min_dice_sim': sim_th,
                     'min_cluster_cos_sim': sim_th, 'download_path': str(download_path)})
    matches = find_matches(docdir, archive_docdir, common_docdir)
    if jsondir:
        if matches:
            try:
                write_doc_pair_matches_to_json(matches, jsondir)
                click.echo(f'Successfully wrote JSON reports to {jsondir}.')
            except NotADirectoryError as e:
                raise UsageError(str(e)) from e
    else:
        text_report = generate_text_report(matches)
        click.echo(f'\n{text_report}')
    sys.exit(0)


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(prog_name='PlagDef', package_name='plagdef')
def gui():
    """
    \b
    PlagDef supports plagiarism detection for student assignments.
    The GUI for this tool is based on the Qt 6 framework and works on all platforms.
    """
    from plagdef.gui.main import MyQtApp
    os.environ["QT_LOGGING_RULES"] = "*.warning=false"
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    app = MyQtApp()
    app.window.show()
    sys.exit(app.exec_())


def find_matches(docdir: tuple, archive_docdir: tuple, common_docdir: tuple) -> list[DocumentPairMatches]:
    try:
        doc_repo = DocumentFileRepository(Path(str(docdir[0])), recursive=docdir[1])
        archive_repo = common_repo = None
        if archive_docdir:
            archive_repo = DocumentFileRepository(
                Path(str(archive_docdir[0])), recursive=archive_docdir[1])
        if common_docdir:
            common_repo = DocumentFileRepository(
                Path(str(common_docdir[0])), recursive=common_docdir[1])
        settings['last_common_docdir'] = common_docdir
        return services.find_matches(doc_repo, archive_repo=archive_repo, common_doc_repo=common_repo)
    except NotADirectoryError as e:
        raise UsageError(str(e)) from e


def reanalyze_pair(doc1: Document, doc2: Document, sim: float):
    ser, last_sim = settings['ser'], settings['min_cos_sim']
    settings.update({'ser': False, 'min_cos_sim': sim, 'min_dice_sim': sim, 'min_cluster_cos_sim': sim})
    common_repo = None
    if 'last_common_docdir' in settings and settings['last_common_docdir']:
        common_repo = DocumentFileRepository(
            Path(str(settings['last_common_docdir'][0])), recursive=settings['last_common_docdir'][1])
    doc_repo = DocumentPairRepository(Document(doc1.name, doc1.path, doc1.text),
                                      Document(doc2.name, doc2.path, doc2.text))
    matches = services.find_matches(doc_repo, common_doc_repo=common_repo, download=False)
    settings.update({'ser': ser, 'min_cos_sim': last_sim, 'min_dice_sim': last_sim, 'min_cluster_cos_sim': last_sim})
    return matches


def write_doc_pair_matches_to_json(matches, jsondir):
    repo = DocumentPairMatchesJsonRepository(Path(str(jsondir)))
    services.write_json_reports(matches, repo)


def read_doc_pair_matches_from_json(jsondir) -> set[DocumentPairMatches]:
    repo = DocumentPairMatchesJsonRepository(Path(str(jsondir)))
    return repo.list()


if __name__ == "__main__":
    args = sys.argv
    if len(args) > 1 and args[1] == 'gui':
        del args[1]
        gui()
    else:
        cli()



from __future__ import annotations

import bz2
import logging
import os
import re
from collections import Counter, defaultdict
from copy import deepcopy
from hashlib import blake2b
from io import BytesIO
from json import JSONDecodeError
from multiprocessing import Lock
from pathlib import Path
from pickle import dump, load, UnpicklingError
from unicodedata import normalize
from urllib.parse import urlparse

import jsonpickle
import magic
import pdfplumber
from magic import MagicException
from ocrmypdf import ocr
from sortedcontainers import SortedSet
from tqdm.contrib.concurrent import process_map

from plagdef.config import settings
from plagdef.model import models

log = logging.getLogger(__name__)
jsonpickle.set_encoder_options('json', indent=4)
lock = Lock()


class FileRepository:
    def __init__(self, base_path: Path, recursive=False):
        self.base_path = base_path
        self._recursive = recursive
        if not base_path.is_dir():
            raise NotADirectoryError(f'The given path {base_path} does not point to an existing directory!')

    def list(self) -> set[models.File]:
        files = list()
        f_gen = self.base_path.rglob('*') if self._recursive else self.base_path.iterdir()
        for f in f_gen:
            if f.is_file():
                binary = not magic.Magic(mime=True).from_buffer(open(f, 'rb').read(2048)).startswith("text")
                try:
                    content = f.read_bytes() if binary else self._read_text(f)
                    files.append(models.File(f, content, binary))
                except UnsupportedFileFormatError as e:
                    log.error(e)
                    log.debug('Following error occurred:', exc_info=True)
        file_groups = defaultdict(list)
        [file_groups[f].append(f) for f in files]
        duplicate_files = list(filter(lambda file_group: len(file_group) > 1, file_groups.values()))
        log.warning(f'Only one representative of the following file groups is included because group members have '
                    f'identical contents: {str(duplicate_files)}') if len(files) != len(file_groups) else None
        return set(file_groups.keys())

    def _read_text(self, file_path: Path):
        try:
            detect_enc = magic.Magic(mime_encoding=True)
            enc = detect_enc.from_buffer(open(file_path, 'rb').read(2048))
            enc_str = enc if enc != 'utf-8' else 'utf-8-sig'
            text = file_path.read_text(encoding=enc_str)
            return normalize('NFC', text)
        except (UnicodeDecodeError, LookupError, MagicException):
            raise UnsupportedFileFormatError(
                f"The file '{file_path.name}' has an unsupported encoding and cannot be read.")

    def save_all(self, files: set[models.File]):
        existing_files = self.list()
        for file in files:
            try:
                self._save(file, existing_files)
                existing_files.add(file)
            except FileExistsError as e:
                log.debug(e)

    def _save(self, file: models.File, existing_files: set[models.File]):
        suffix = 1
        saved = False
        while not saved:
            if file in existing_files:
                raise FileExistsError(f'File "{file.path.name} was not saved because there already is a file with '
                                      f'identical contents.')
            if file.path.exists():
                file.path = file.path.with_name(f'{file.path.stem.replace(f"_{suffix - 1}", "")}_{suffix}'
                                                f'{file.path.suffix}')
                suffix = suffix + 1
                continue
            if file.binary:
                with file.path.open('wb') as f:
                    f.write(file.content)
            else:
                with file.path.open('w', encoding="utf-8") as f:
                    f.write(file.content)
            saved = True

    def remove_all(self, files: set[models.File]):
        for file in files:
            file.path.unlink(missing_ok=True)


class DocumentFileRepository:
    def __init__(self, dir_path: Path, recursive=False, lang=None, use_ocr=None):
        self._file_repo = FileRepository(dir_path, recursive)
        self.lang = lang if lang else settings['lang']
        self._use_ocr = use_ocr if use_ocr else settings['ocr']

    @property
    def base_path(self):
        return self._file_repo.base_path

    def list(self) -> set[models.Document]:
        files = list(filter(lambda f: not f.path.suffix.lower() == '.pdef', self._file_repo.list()))
        docs = process_map(self._create_doc, files, desc=f"Reading documents in '{self.base_path}'",
                           unit='doc', total=len(files), max_workers=os.cpu_count())
        return set(filter(None, docs))

    def _create_doc(self, file: models.File) -> models.Document:
        if file.path.suffix.lower() == '.pdf':
            reader = PdfReader(file.path, self.lang, self._use_ocr)
            text = reader.extract_text()
            doc = models.Document(file.path.stem, str(file.path), text)
            urls = reader.extract_urls()
            doc.urls.update(urls) if urls else None
        elif not file.binary:
            doc = models.Document(file.path.stem, str(file.path), file.content)
        else:
            log.error(f'The file "{file.path.name}" has an unsupported encoding and cannot be read.')
            doc = None
        return doc

    def save_all(self, docs: set[models.Document]):
        files = {models.File(Path(doc.path), doc.text, False) for doc in docs}
        self._file_repo.save_all(files)

    def remove_all(self, docs: set[models.Document]):
        files = {models.File(Path(doc.path), doc.text, False) for doc in docs}
        self._file_repo.remove_all(files)


class DocumentPairRepository:
    def __init__(self, doc1: models.Document, doc2: models.Document, lang=None):
        self._docs = {doc1, doc2}
        self.lang = lang if lang else settings['lang']

    def list(self) -> set[models.Document]:
        return self._docs


class DocumentPairMatchesJsonRepository:
    def __init__(self, out_path: Path):
        if not out_path.is_dir():
            raise NotADirectoryError(f"The given path '{out_path}' does not point to an existing directory!")
        self._out_path = out_path

    def save(self, doc_pair_matches):
        clone = deepcopy(doc_pair_matches)
        clone.doc1.vocab, clone.doc2.vocab = Counter(), Counter()
        clone.doc1._sents, clone.doc2._sents = SortedSet(), SortedSet()
        file_name = Path(f'{clone.doc1.name}-{clone.doc2.name}.json')
        file_path = self._out_path / file_name
        with file_path.open('w', encoding='utf-8') as f:
            text = jsonpickle.encode(clone)
            f.write(text)

    def list(self) -> set[models.DocumentPairMatches]:
        doc_pair_matches_list = set()
        for file in self._out_path.iterdir():
            if file.is_file() and file.suffix == '.json':
                try:
                    text = file.read_text(encoding='utf-8')
                    doc_pair_matches = jsonpickle.decode(text)
                    doc_pair_matches_list.add(doc_pair_matches)
                except (UnicodeDecodeError, JSONDecodeError):
                    log.error(f"The file '{file.name}' could not be read.")
                    log.debug('Following error occurred:', exc_info=True)
        return doc_pair_matches_list


class DocumentPickleRepository:
    def __init__(self, dir_path: Path, common_dir_path: Path = None):
        if not dir_path.is_dir():
            raise NotADirectoryError(f"The given path '{dir_path}' does not point to an existing directory!")
        path_hash = blake2b(str(common_dir_path).encode(), digest_size=16).hexdigest()
        self.file_path = dir_path / f'.{path_hash}.pdef'

    def save(self, docs: set[models.Document]):
        with bz2.open(self.file_path, 'wb') as file:
            dump(docs, file)

    def list(self) -> set[models.Document]:
        if self.file_path.exists():
            log.info('Found preprocessing file. Deserializing...')
            try:
                with bz2.open(self.file_path, 'rb') as file:
                    return load(file)
            except (UnpicklingError, EOFError):
                log.warning(f"Could not deserialize preprocessing file, '{self.file_path.name}' seems to be corrupted.")
                log.debug('Following error occurred:', exc_info=True)
        return set()


class PdfReader:
    ERROR_HEURISTIC = '¨[aou]|ﬀ|\(cid:\d+\)|[a-zA-Z]{50}'

    def __init__(self, file, lang, use_ocr):
        self._file = file
        self._lang = 'eng' if lang == 'en' else 'deu'
        self._use_ocr = use_ocr

    def extract_urls(self) -> set[str]:
        # Temporary fix for: https://github.com/jsvine/pdfplumber/issues/463
        try:
            with pdfplumber.open(self._file) as pdf:
                parsed_urls = {urlparse(uri_obj['uri'], "https") for uri_obj in pdf.hyperlinks}
                return {parsed_url.geturl().rstrip('/').replace("///", "//")
                        for parsed_url in filter(lambda url: url.scheme in ("http", "https"), parsed_urls)}
        except UnicodeDecodeError:
            log.warning(f'Could not extract hyperlinks from PDF "{self._file.name}".')

    def extract_text(self):
        text = self._extract()
        if self._use_ocr and self._poor_extraction(text):
            log.warning(f"Poor text extraction in '{self._file.name}' detected! Using OCR...")
            with BytesIO() as ocr_file:
                with lock:
                    ocr(self._file, ocr_file, language=self._lang, force_ocr=True, progress_bar=False,
                        max_image_mpixels=512)
                text = self._extract(ocr_file)
        return text

    def _extract(self, file=None) -> str:
        if file is None:
            file = self._file
        with pdfplumber.open(file) as pdf:
            text = ' '.join(filter(None, (page.extract_text() for page in pdf.pages)))
            normalized_text = normalize('NFC', text)
            return re.sub('-\s?\n', '', normalized_text)  # Merge hyphenated words

    def _poor_extraction(self, text: str) -> bool:
        return not len(text.strip()) or bool(re.search(PdfReader.ERROR_HEURISTIC, text))


class UnsupportedFileFormatError(Exception):
    pass



from __future__ import annotations

import sys
from pathlib import Path

import pkg_resources
from PySide6.QtCore import QFile, QIODevice, Qt
from PySide6.QtGui import QCursor, QMovie
from PySide6.QtUiTools import QUiLoader
from PySide6.QtWidgets import QButtonGroup, QMainWindow, QFileDialog, QDialog

from plagdef.config import settings
from plagdef.gui.model import ResultsTableModel, DocumentPairMatches
from plagdef.model import models
from plagdef.util import version, truncate

UI_FILES = {
    'main_window': pkg_resources.resource_filename(__name__, 'ui/main_window.ui'),
    'home_widget': pkg_resources.resource_filename(__name__, 'ui/home_widget.ui'),
    'loading_widget': pkg_resources.resource_filename(__name__, 'ui/loading_widget.ui'),
    'error_widget': pkg_resources.resource_filename(__name__, 'ui/error_widget.ui'),
    'no_results_widget': pkg_resources.resource_filename(__name__, 'ui/no_results_widget.ui'),
    'results_widget': pkg_resources.resource_filename(__name__, 'ui/results_widget.ui'),
    'matches_dialog': pkg_resources.resource_filename(__name__, 'ui/matches_dialog.ui'),
    'msg_dialog': pkg_resources.resource_filename(__name__, 'ui/msg_dialog.ui'),
    'settings_dialog': pkg_resources.resource_filename(__name__, 'ui/settings_dialog.ui'),
}


class MainWindow:
    def __init__(self, views: list[View]):
        self._window = _load_ui_file(Path(UI_FILES['main_window']))
        self._views = views
        self._configure()

    def _configure(self):
        self._window.setWindowTitle(f'PlagDef v{version()}')
        for view in self._views:
            self._window.stacked_widget.addWidget(view.widget)

    def switch_to(self, view_cls: type, data=None):
        idx = self._window.stacked_widget.currentIndex()
        for view in self._views:
            if type(view) == view_cls:
                self._views[idx].on_destroy()
                self._window.stacked_widget.setCurrentWidget(view.widget)
                view.on_init(data)

    def show(self):
        self._window.show()


class View:
    def on_init(self, data=None):
        pass

    def on_destroy(self):
        pass


class HomeView(View):
    def __init__(self, lang=settings['lang']):
        self.widget = _load_ui_file(Path(UI_FILES['home_widget']))
        self._configure(lang)

    def _configure(self, lang):
        self.widget.lang_button_group = QButtonGroup()
        self.widget.lang_button_group.addButton(self.widget.ger_button)
        self.widget.lang_button_group.addButton(self.widget.eng_button)
        self.widget.ger_button.click() if lang == 'de' else self.widget.eng_button.click()
        [element.setVisible(False)
         for element in (self.widget.archive_dir_label, self.widget.archive_rmdir_button, self.widget.docs_dir_label,
                         self.widget.docs_rmdir_button, self.widget.common_dir_label, self.widget.common_rmdir_button)]
        [button.setCursor(QCursor(Qt.PointingHandCursor))
         for button in (self.widget.ger_button, self.widget.eng_button, self.widget.archive_dir_button,
                        self.widget.docs_dir_button, self.widget.common_dir_button, self.widget.detect_button,
                        self.widget.archive_rmdir_button, self.widget.docs_rmdir_button,
                        self.widget.common_rmdir_button, self.widget.open_button, self.widget.settings_button)]

    @property
    def archive_rec(self):
        return self.widget.archive_rec_check_box.isChecked()

    @property
    def docs_rec(self):
        return self.widget.docs_rec_check_box.isChecked()

    @property
    def common_rec(self):
        return self.widget.common_rec_check_box.isChecked()

    @property
    def lang(self):
        return 'de' if self.widget.lang_button_group.checkedButton() == self.widget.ger_button else 'en'

    def register_for_signals(self, select_lang=None, open_report_dir=None, select_archive_dir=None, rm_archive_dir=None,
                             select_docs_dir=None, rm_docs_dir=None, select_common_dir=None, rm_common_dir=None,
                             detect=None, settings=None):
        self.widget.lang_button_group.buttonClicked.connect(lambda: select_lang())
        self.widget.settings_button.clicked.connect(lambda: settings())
        self.widget.open_button.clicked.connect(lambda: open_report_dir())
        self.widget.archive_dir_button.clicked.connect(lambda: select_archive_dir())
        self.widget.archive_rmdir_button.clicked.connect(lambda: rm_archive_dir())
        self.widget.docs_dir_button.clicked.connect(lambda: select_docs_dir())
        self.widget.docs_rmdir_button.clicked.connect(lambda: rm_docs_dir())
        self.widget.common_dir_button.clicked.connect(lambda: select_common_dir())
        self.widget.common_rmdir_button.clicked.connect(lambda: rm_common_dir())
        self.widget.detect_button.clicked.connect(lambda: detect())

    def archive_dir_selected(self, folder_name: str):
        self._dir_selected(folder_name, self.widget.archive_dir_button, self.widget.archive_rmdir_button,
                           self.widget.archive_dir_label, 12)

    def archive_dir_removed(self):
        self._dir_removed(self.widget.archive_dir_button, self.widget.archive_rmdir_button,
                          self.widget.archive_dir_label)

    def docs_dir_selected(self, folder_name: str):
        self._dir_selected(folder_name, self.widget.docs_dir_button, self.widget.docs_rmdir_button,
                           self.widget.docs_dir_label, 16)
        self.widget.detect_button.setEnabled(True)

    def docs_dir_removed(self):
        self._dir_removed(self.widget.docs_dir_button, self.widget.docs_rmdir_button, self.widget.docs_dir_label)
        self.widget.detect_button.setEnabled(False)

    def common_dir_selected(self, folder_name: str):
        self._dir_selected(folder_name, self.widget.common_dir_button, self.widget.common_rmdir_button,
                           self.widget.common_dir_label, 12)

    def common_dir_removed(self):
        self._dir_removed(self.widget.common_dir_button, self.widget.common_rmdir_button, self.widget.common_dir_label)

    def _dir_selected(self, folder_name: str, dir_button, rm_button, label, font_size: int):
        dir_button.setEnabled(False)
        label.setText(
            f'<html><head/><body><p align="center"><span style="font-size:{font_size}pt; color:#ffffff;"> '
            f'{folder_name}</span></p></body></html>')
        label.setVisible(True)
        rm_button.setVisible(True)

    def _dir_removed(self, button, rm_button, label):
        label.setVisible(False)
        rm_button.setVisible(False)
        button.setEnabled(True)

    def on_destroy(self):
        self.archive_dir_removed()
        self.docs_dir_removed()
        self.common_dir_removed()


class FileDialog(QFileDialog):
    def __init__(self):
        super().__init__()
        self.setFileMode(QFileDialog.Directory)
        self.selected_dir = None

    def open(self) -> bool:
        if self.exec_() == QDialog.Accepted:
            self.selected_dir = self.selectedFiles()[0]
            return True


class LoadingView(View):
    def __init__(self):
        self._movie = QMovie(':/loading.gif')
        self.widget = _load_ui_file(Path(UI_FILES['loading_widget']))
        self._configure()

    def _configure(self):
        self.widget.loading_movie_label.setMovie(self._movie)

    def on_init(self, data=None):
        self._movie.start()

    def on_destroy(self):
        self._movie.stop()


class NoResultsView(View):
    def __init__(self):
        self.widget = _load_ui_file(Path(UI_FILES['no_results_widget']))
        self._configure()

    def _configure(self):
        self.widget.again_button_no_res.setCursor(QCursor(Qt.PointingHandCursor))

    def register_for_signals(self, again=None):
        self.widget.again_button_no_res.clicked.connect(lambda: again())


class ErrorView(View):
    def __init__(self):
        self.widget = _load_ui_file(Path(UI_FILES['error_widget']))
        self._configure()

    def _configure(self):
        self.widget.again_button_err.setCursor(QCursor(Qt.PointingHandCursor))

    def on_init(self, data=None):
        self.widget.error_msg_label.setText(
            f'<html><head/><body><p align="center"><span style=" font-size:12pt; color:#ffffff;"> '
            f'{data}</span></p></body></html>')
        self.widget.error_msg_label.setWordWrap(True)

    def register_for_signals(self, again=None):
        self.widget.again_button_err.clicked.connect(lambda: again())


class ResultView(View):
    def __init__(self):
        self.widget = _load_ui_file(Path(UI_FILES['results_widget']))
        self._configure()
        self.doc_pair_matches = None

    def _configure(self):
        self.widget.export_button.setCursor(QCursor(Qt.PointingHandCursor))
        self.widget.again_button_res.setCursor(QCursor(Qt.PointingHandCursor))

    def on_init(self, data=None):
        self.doc_pair_matches = data
        self.widget.doc_pairs_label.setText(f"Found {len(data) if len(data) else 'no'} suspicious document pair"
                                            f"{'s' if len(data) > 1 else ''}.")
        self._set_table_model(self.widget.verbatim_results, models.MatchType.VERBATIM, data)
        self._set_table_model(self.widget.intelligent_results, models.MatchType.INTELLIGENT, data)
        self._set_table_model(self.widget.summary_results, models.MatchType.SUMMARY, data)
        self._hide_empty_tables()

    def _set_table_model(self, table, match_type, pairs):
        table.setModel(ResultsTableModel(match_type, pairs))
        table.resizeRowsToContents()
        table.resizeColumnsToContents()
        table.parentWidget().adjustSize()

    def _hide_empty_tables(self):
        self.widget.doc_pair_tabs.setTabVisible(0, bool(self.widget.verbatim_results.model().rowCount()))
        self.widget.doc_pair_tabs.setTabVisible(1, bool(self.widget.intelligent_results.model().rowCount()))
        self.widget.doc_pair_tabs.setTabVisible(2, bool(self.widget.summary_results.model().rowCount()))

    @property
    def selected_matches(self) -> list[models.DocumentPairMatches]:
        matches = []
        verbatim_indices = self.widget.verbatim_results.selectionModel().selectedIndexes()
        verbatim_indices = filter(lambda i: i.column() == 0, verbatim_indices)
        for index in verbatim_indices:
            verbatim_matches = self.widget.verbatim_results.model().doc_pair_matches(index)
            matches.extend(verbatim_matches.matches)
        intelligent_indices = self.widget.intelligent_results.selectionModel().selectedIndexes()
        intelligent_indices = filter(lambda i: i.column() == 0, intelligent_indices)
        for index in intelligent_indices:
            intelligent_matches = self.widget.intelligent_results.model().doc_pair_matches(index)
            matches.extend(intelligent_matches.matches)
        summary_indices = self.widget.summary_results.selectionModel().selectedIndexes()
        summary_indices = filter(lambda i: i.column() == 0, summary_indices)
        for index in summary_indices:
            summary_matches = self.widget.summary_results.model().doc_pair_matches(index)
            matches.extend(summary_matches.matches)
        return models.DocumentPairMatches.from_matches(matches)

    def register_for_signals(self, export=None, again=None, select_pair=None):
        self.widget.export_button.clicked.connect(lambda: export())
        self.widget.again_button_res.clicked.connect(lambda: again())
        self.widget.verbatim_results.doubleClicked.connect(
            lambda index: select_pair(index.model().doc_pair_matches(index)))
        self.widget.intelligent_results.doubleClicked.connect(
            lambda index: select_pair(index.model().doc_pair_matches(index)))
        self.widget.summary_results.doubleClicked.connect(
            lambda index: select_pair(index.model().doc_pair_matches(index)))


class MatchesDialog:
    def __init__(self):
        self.widget = _load_ui_file(Path(UI_FILES['matches_dialog']))
        self._configure()
        self._doc_pair_matches = None
        self._selected = 0

    def _configure(self):
        self.widget.doc1_label.setCursor(QCursor(Qt.PointingHandCursor))
        self.widget.doc2_label.setCursor(QCursor(Qt.PointingHandCursor))
        self.widget.reanalyze_button.setCursor(QCursor(Qt.PointingHandCursor))
        self.widget.sim_slider.valueChanged.connect(lambda val: self._update_label(val))

    def _update_label(self, sim):
        self.widget.value_label.setText(str(sim / 10))

    @property
    def doc1(self):
        return self._doc_pair_matches.doc1

    @property
    def doc2(self):
        return self._doc_pair_matches.doc2

    @property
    def sim_threshold(self) -> float:
        return float(self.widget.value_label.text())

    @property
    def match_type(self):
        return self._doc_pair_matches.matches[0].type

    def open(self, doc_pair_matches: DocumentPairMatches):
        self.widget.value_label.setText(str(settings['min_cos_sim']))
        self.widget.sim_slider.setValue(settings['min_cos_sim'] * 10)
        self.set_data(doc_pair_matches)
        self.widget.exec_()

    def set_data(self, doc_pair_matches):
        self._doc_pair_matches = doc_pair_matches
        self.widget.doc1_label.setText(truncate(self._doc_pair_matches.doc1.name, 50))
        self.widget.doc1_path.setText(f'({truncate(self._doc_pair_matches.doc1.path, 65)})')
        self.widget.doc2_label.setText(truncate(self._doc_pair_matches.doc2.name, 50))
        self.widget.doc2_path.setText(f'({truncate(self._doc_pair_matches.doc2.path, 65)})')
        self._selected = 0
        self._show_match()

    def _show_match(self):
        doc1, doc2 = self._doc_pair_matches.doc1, self._doc_pair_matches.doc2
        self.widget.doc1_text.setText(self._doc_pair_matches.matches[self._selected].frag_from_doc(doc1).text)
        self.widget.doc2_text.setText(self._doc_pair_matches.matches[self._selected].frag_from_doc(doc2).text)
        self.widget.page_label.setText(f'{self._selected + 1}/{len(self._doc_pair_matches)}')
        self.widget.prev_button.setVisible(True)
        self.widget.next_button.setVisible(True)
        if self._selected == 0:
            self.widget.prev_button.setVisible(False)
        if self._selected + 1 == len(self._doc_pair_matches):
            self.widget.next_button.setVisible(False)

    def register_for_signals(self, prev_match=None, next_match=None, open_doc=None, reanalyze=None):
        self.widget.prev_button.clicked.connect(lambda: prev_match())
        self.widget.next_button.clicked.connect(lambda: next_match())
        self.widget.doc1_label.clicked.connect(lambda: open_doc(self._doc_pair_matches.doc1.path))
        self.widget.doc2_label.clicked.connect(lambda: open_doc(self._doc_pair_matches.doc2.path))
        self.widget.reanalyze_button.clicked.connect(lambda: reanalyze())

    def prev_match(self):
        self._selected -= 1
        self._show_match()

    def next_match(self):
        self._selected += 1
        self._show_match()

    def reanalyzing(self, loading: bool):
        self.widget.sim_slider.setEnabled(not loading)
        self.widget.reanalyze_button.setEnabled(not loading)


class MessageDialog:
    def __init__(self, text):
        self.widget = _load_ui_file(Path(UI_FILES['msg_dialog']))
        self.widget.msg_label.setText(text)
        self.widget.exec_()


class SettingsDialog:
    def __init__(self):
        self.widget = _load_ui_file(Path(UI_FILES['settings_dialog']))
        self.widget.sim_slider.valueChanged.connect(lambda val: self._update_label(val))

    def _update_label(self, sim):
        self.widget.value_label.setText(str(sim / 10))

    @property
    def similarity_threshold(self) -> float:
        return float(self.widget.value_label.text())

    @property
    def ocr(self) -> bool:
        return self.widget.ocr_check_box.isChecked()

    @property
    def download_path(self) -> bool:
        return self.widget.es_line_edit.text()

    def register_for_signals(self, select_download_path=None):
        self.widget.es_button.clicked.connect(lambda: select_download_path())

    def open(self, ocr=None, sim=None, download_path=None):
        ocr = settings['ocr'] if not ocr else ocr
        sim = settings['min_cos_sim'] if not sim else sim
        download_path = settings['download_path'] if not download_path else download_path
        self.widget.ocr_check_box.setChecked(ocr)
        self.widget.es_line_edit.setText(download_path)
        self.widget.value_label.setText(str(sim))
        self.widget.sim_slider.setValue(sim * 10)
        self.widget.exec_()

    def download_dir_selected(self, dir_path):
        self.widget.es_button.setText("Open..." if not dir_path else "Clear")
        self.widget.es_line_edit.setText(dir_path)


def _load_ui_file(path: Path) -> QMainWindow:
    main_ui = QFile(str(path))
    if not main_ui.open(QIODevice.ReadOnly):
        print(f"Cannot open {path}: {main_ui.errorString()}")
        sys.exit(-1)
    loader = QUiLoader()
    window = loader.load(main_ui)
    main_ui.close()
    if not window:
        print(loader.errorString())
        sys.exit(-1)
    return window



from __future__ import annotations

import os
import platform
import subprocess

from click import UsageError

import plagdef.gui.main as main
from plagdef.app import write_doc_pair_matches_to_json, read_doc_pair_matches_from_json
from plagdef.config import settings
from plagdef.gui.model import DocumentPairMatches
from plagdef.gui.views import HomeView, LoadingView, NoResultsView, ErrorView, ResultView, \
    FileDialog, MatchesDialog, MessageDialog, SettingsDialog
from plagdef.model import models


class HomeController:
    def __init__(self):
        self.view = HomeView()
        self.settings_controller = SettingsController()
        self.archive_dir_dialog = FileDialog()
        self.docs_dir_dialog = FileDialog()
        self.common_dir_dialog = FileDialog()
        self._connect_slots()

    def _connect_slots(self):
        self.view.register_for_signals(select_lang=self.on_select_lang,
                                       open_report_dir=self.on_open_click,
                                       select_archive_dir=self._on_select_archive_dir,
                                       rm_archive_dir=self._on_remove_archive_dir,
                                       select_docs_dir=self._on_select_docs_dir,
                                       rm_docs_dir=self._on_remove_docs_dir,
                                       select_common_dir=self._on_select_common_dir,
                                       rm_common_dir=self._on_remove_common_dir,
                                       detect=self._on_detect,
                                       settings=self._on_settings_click)

    def on_select_lang(self):
        settings.update({'lang': self.view.lang})

    def on_open_click(self):
        dialog = FileDialog()
        if dialog.open():
            matches = read_doc_pair_matches_from_json(dialog.selected_dir)
            if len(matches):
                main.app.window.switch_to(ResultView, matches)
            else:
                MessageDialog('The selected folder contains no match files.')

    def _on_settings_click(self):
        self.settings_controller.view.open()
        settings.update({'ocr': self.settings_controller.view.ocr,
                         'download_path': self.settings_controller.view.download_path,
                         'min_cos_sim': self.settings_controller.view.similarity_threshold,
                         'min_dice_sim': self.settings_controller.view.similarity_threshold,
                         'min_cluster_cos_sim': self.settings_controller.view.similarity_threshold})

    def _on_select_archive_dir(self):
        if self.archive_dir_dialog.open():
            folder_name = self.archive_dir_dialog.selected_dir[self.archive_dir_dialog.selected_dir.rfind("/"):]
            self.view.archive_dir_selected(folder_name)

    def _on_remove_archive_dir(self):
        self.archive_dir_dialog.selected_dir = None
        self.view.archive_dir_removed()

    def _on_select_docs_dir(self):
        if self.docs_dir_dialog.open():
            folder_name = self.docs_dir_dialog.selected_dir[self.docs_dir_dialog.selected_dir.rfind("/"):]
            self.view.docs_dir_selected(folder_name)

    def _on_remove_docs_dir(self):
        self.docs_dir_dialog.selected_dir = None
        self.view.docs_dir_removed()

    def _on_select_common_dir(self):
        if self.common_dir_dialog.open():
            folder_name = self.common_dir_dialog.selected_dir[self.common_dir_dialog.selected_dir.rfind("/"):]
            self.view.common_dir_selected(folder_name)

    def _on_remove_common_dir(self):
        self.common_dir_dialog.selected_dir = None
        self.view.common_dir_removed()

    def _on_detect(self):
        doc_dir = (self.docs_dir_dialog.selected_dir, self.view.docs_rec) if self.docs_dir_dialog.selected_dir else None
        archive_dir = (self.archive_dir_dialog.selected_dir, self.view.archive_rec) \
            if self.archive_dir_dialog.selected_dir else None
        common_dir = (self.common_dir_dialog.selected_dir, self.view.common_rec) \
            if self.common_dir_dialog.selected_dir else None
        main.app.find_matches(doc_dir, archive_dir, common_dir, self._on_detect_success, self._on_detect_error)
        main.app.window.switch_to(LoadingView)
        self.archive_dir_dialog.selected_dir = self.common_dir_dialog.selected_dir \
            = self.docs_dir_dialog.selected_dir = None

    def _on_detect_success(self, matches: list[models.DocumentPairMatches]):
        if matches:
            main.app.window.switch_to(ResultView, matches)
        else:
            main.app.window.switch_to(NoResultsView)

    def _on_detect_error(self, error: (type, Exception)):
        if error[0] == UsageError:
            main.app.window.switch_to(ErrorView, str(error[1]))
        else:
            main.app.window.switch_to(ErrorView,
                                      'An error occurred. Please refer to the command line for more details.')
            raise error[1]


class SettingsController:
    def __init__(self):
        self.view = SettingsDialog()
        self._download_dir_dialog = FileDialog()
        self.view.register_for_signals(self._on_download_path_select)

    def _on_download_path_select(self):
        if self.view.download_path:
            self.view.download_dir_selected(None)
        elif self._download_dir_dialog.open():
            self.view.download_dir_selected(self._download_dir_dialog.selected_dir)


class LoadingController:
    def __init__(self):
        self.view = LoadingView()


class NoResultsController:
    def __init__(self):
        self.view = NoResultsView()
        self._connect_slots()

    def _connect_slots(self):
        self.view.register_for_signals(self._on_again)

    def _on_again(self):
        main.app.window.switch_to(HomeView)


class ErrorController:
    def __init__(self):
        self.view = ErrorView()
        self._connect_slots()

    def _connect_slots(self):
        self.view.register_for_signals(self._on_again)

    def _on_again(self):
        main.app.window.switch_to(HomeView)


class ResultController:
    def __init__(self):
        self.view = ResultView()
        self.matches_dialog = MatchesDialog()
        self._connect_slots()

    def _connect_slots(self):
        self.view.register_for_signals(self.on_export, self._on_again, self.on_select_pair)
        self.matches_dialog.register_for_signals(self.on_prev_match, self.on_next_match, self.on_doc_name_click,
                                                 self.on_reanalyze_click)

    def on_export(self):
        selected_matches = self.view.selected_matches
        if not selected_matches:
            MessageDialog(f"Please select matches first!")
        else:
            dialog = FileDialog()
            if dialog.open():
                write_doc_pair_matches_to_json(selected_matches, dialog.selected_dir)
                MessageDialog(f"Successfully generated {len(selected_matches)} JSON report(s).")

    def _on_again(self):
        main.app.window.switch_to(HomeView)

    def on_select_pair(self, doc_pair_matches: DocumentPairMatches):
        self.matches_dialog.open(doc_pair_matches)

    def on_prev_match(self):
        self.matches_dialog.prev_match()

    def on_next_match(self):
        self.matches_dialog.next_match()

    def on_doc_name_click(self, path: str):
        if platform.system() == 'Darwin':  # macOS
            subprocess.call(('open', path))
        elif platform.system() == 'Windows':  # Windows
            os.startfile(path)
        else:  # linux variants
            subprocess.call(('xdg-open', path))

    def on_reanalyze_click(self):
        self.matches_dialog.reanalyzing(True)
        main.app.reanalyze_pair(self.matches_dialog.doc1, self.matches_dialog.doc2,
                                self.matches_dialog.sim_threshold, self._on_reanalyze_success, self._on_reanalyze_error)

    def _on_reanalyze_success(self, matches: list[models.DocumentPairMatches]):
        self.matches_dialog.reanalyzing(False)
        if matches:
            self.matches_dialog.set_data(DocumentPairMatches.from_model(matches[0], self.matches_dialog.match_type))
        else:
            MessageDialog("No matches found.")

    def _on_reanalyze_error(self, error: (type, Exception)):
        self.matches_dialog.reanalyzing(False)
        if error[0] == UsageError:
            MessageDialog(str(error[1]))
        else:
            MessageDialog('An error occurred. Please refer to the command line for more details.')
            raise error[1]



from __future__ import annotations

from dataclasses import dataclass

from PySide6.QtCore import QAbstractTableModel, QModelIndex, Qt
from PySide6.QtGui import QColor

from plagdef.model import models
from plagdef.model.models import MatchType
from plagdef.util import truncate


@dataclass(frozen=True)
class DocumentPairMatches:
    doc1: models.Document
    doc2: models.Document
    matches: list[models.Match]

    @classmethod
    def from_model(cls, model: models.DocumentPairMatches, match_type) -> DocumentPairMatches:
        typed_matches = model.list(match_type)
        if len(typed_matches):
            d1 = model.doc1 if model.doc1.name < model.doc2.name else model.doc2
            d2 = model.doc2 if d1 == model.doc1 else model.doc1
            return DocumentPairMatches(d1, d2,
                                       sorted(typed_matches,
                                              key=lambda m, doc1=d1: m.frag_from_doc(doc1).start_char))

    def __len__(self):
        return len(self.matches)


class ResultsTableModel(QAbstractTableModel):
    def __init__(self, match_type: MatchType, doc_pair_matches: list[models.DocumentPairMatches]):
        super().__init__()
        self._doc_pair_matches = [DocumentPairMatches.from_model(matches, match_type) for matches in doc_pair_matches]
        self._doc_pair_matches = sorted(filter(None, self._doc_pair_matches), key=lambda m: m.doc1.name)

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = Qt.DisplayRole):
        headers = ['Document 1', 'Document 2']
        if role == Qt.DisplayRole and orientation == Qt.Horizontal:
            return headers[section]

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return 2

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:
        return len(self._doc_pair_matches)

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole):
        doc_pair = self._doc_pair_matches[index.row()].doc1, self._doc_pair_matches[index.row()].doc2
        if role == Qt.DisplayRole:
            name = doc_pair[index.column()].name
            return truncate(name, 50)
        elif role == Qt.ForegroundRole:
            return QColor(255, 255, 255)
        elif role == Qt.ToolTipRole:
            return doc_pair[index.column()].path

    def doc_pair_matches(self, index: QModelIndex) -> DocumentPairMatches:
        return self._doc_pair_matches[index.row()]



from __future__ import annotations

import logging
import os
from functools import partial
from mimetypes import guess_extension
from pathlib import Path
from urllib.parse import urlparse

import requests
from bs4 import BeautifulSoup
from requests.exceptions import RequestException
from tqdm.contrib.concurrent import thread_map
from werkzeug.utils import secure_filename

from plagdef.config import settings
from plagdef.model.models import Document, File

log = logging.getLogger(__name__)


def download_all_external_sources(docs: set[Document], target_dir: Path) -> set[File]:
    urls = {url for doc in docs for url in doc.urls}
    files = filter(None, thread_map(partial(_download_page, target_dir=target_dir), urls, max_workers=os.cpu_count(),
                                    total=len(urls), desc='Downloading', unit='external sources'))
    return set(files)


def download_external_sources(doc: Document, target_dir: Path) -> set[File]:
    files = filter(None, thread_map(partial(_download_page, target_dir=target_dir), doc.urls,
                                    max_workers=os.cpu_count(), total=len(doc.urls), desc='Downloading',
                                    unit='external sources'))
    return set(files)


def _download_page(url: str, target_dir: Path) -> File:
    parsed_url = urlparse(url)
    url_path = parsed_url.path.rstrip("/")
    filename = parsed_url.netloc if not url_path else f'{url_path[url_path.rindex("/") + 1:]}_from_{parsed_url.netloc}'
    filename = secure_filename(filename)
    try:
        resp = requests.get(url)
        resp.raise_for_status()
        mime_type = resp.headers['content-type'] if 'content-type' in resp.headers else 'text/html'
        binary = not mime_type.startswith('text')
        content = resp.content if binary else resp.text
        ext = guess_extension(mime_type)
        if mime_type.startswith("text/html"):
            content = BeautifulSoup(resp.text, features="lxml").get_text(" ", True)
            ext = ".txt"
        file = File(Path(f"{target_dir}/{filename}{ext}"), content, binary) \
            if len(content) >= settings['min_ext_size'] else None
        return file
    except RequestException:
        log.debug(f'Could not download from "{url}".')



from __future__ import annotations

from plagdef.model.models import Seed, Cluster


class ClusterBuilder:
    def __init__(self, adjacent_sents_gap: int, min_adjacent_sents_gap: int, min_sent_number: int,
                 min_cluster_cos_sim: float):
        self._adjacent_sents_gap = adjacent_sents_gap
        self._min_adjacent_sents_gap = min_adjacent_sents_gap
        self._min_sent_number = min_sent_number
        self._min_cluster_cos_sim = min_cluster_cos_sim

    def extend(self, seeds: set[Seed], adjacent_sents_gap: int = None) -> set[Cluster]:
        if adjacent_sents_gap is None:
            adjacent_sents_gap = self._adjacent_sents_gap
        clusters = _build_clusters(seeds, adjacent_sents_gap)
        return self._validate(clusters, adjacent_sents_gap)

    def _validate(self, clusters: set[Cluster], adjacent_sents_gap: int) -> set[Cluster]:
        valid_clusters = set()
        for cluster in clusters:
            if cluster.cos_sim > self._min_cluster_cos_sim:
                valid_clusters.add(cluster)
            elif adjacent_sents_gap > self._min_adjacent_sents_gap:
                cluster_detections = self.extend(set(cluster.seeds), adjacent_sents_gap - 1)
                valid_clusters.update(cluster_detections)
        return valid_clusters


def _build_clusters(seeds: set[Seed], adjacent_sents_gap: int) -> set[Cluster]:
    doc1_clusters = _join_seeds(seeds, adjacent_sents_gap, first=True)
    clusters = set()
    for cluster in doc1_clusters:
        clusters.update(_join_seeds(cluster.seeds, adjacent_sents_gap, first=False))
    return clusters


def _join_seeds(seeds: set[Seed], adjacent_sents_gap: int, first: bool) -> set[Cluster]:
    sorted_seeds = sorted(seeds, key=lambda s: s.sent1.start_char if first else s.sent2.start_char)
    clusters = set()
    seed_iter: enumerate = enumerate(sorted_seeds)
    for seed_idx, seed in seed_iter:
        cluster_seeds = [seed]  # Cluster contains at least first seed
        for adj_seed in sorted_seeds[seed_idx + 1:]:  # For following seeds
            sent1 = cluster_seeds[-1].sent1 if first else cluster_seeds[-1].sent2
            sent2 = adj_seed.sent1 if first else adj_seed.sent2
            if sent1.adjacent_to(sent2, adjacent_sents_gap):  # If sent is adjacent, add seed
                cluster_seeds.append(next(seed_iter)[1])
            else:
                break  # Seeds are sorted by sent start_char
        clusters.add(Cluster(set(cluster_seeds)))
    return clusters



from __future__ import annotations

import os
import signal
import sys
import traceback

from PySide6.QtCore import QRunnable, Slot, QObject, Signal, QThreadPool
from PySide6.QtWidgets import QApplication

# noinspection PyUnresolvedReferences
import plagdef.gui.resources
from plagdef.app import find_matches, reanalyze_pair
from plagdef.gui.controllers import HomeController, LoadingController, ErrorController, NoResultsController, \
    ResultController
from plagdef.gui.views import MainWindow

app = None


class MyQtApp(QApplication):
    def __init__(self):
        super().__init__()
        self.aboutToQuit.connect(lambda: os.kill(os.getpid(), signal.SIGINT))
        self._ctrls = [HomeController(), LoadingController(), ErrorController(), NoResultsController(),
                       ResultController()]
        self._views = [ctrl.view for ctrl in self._ctrls]
        self.window = MainWindow(self._views)
        global app
        app = self

    def find_matches(self, docdir: tuple[str, bool], archive_docdir: [str, bool],
                     common_docdir: [str, bool], on_success, on_error):
        worker = Worker(find_matches, docdir, archive_docdir, common_docdir)
        worker.signals.result.connect(on_success)
        worker.signals.error.connect(on_error)
        pool = QThreadPool.globalInstance()
        pool.start(worker)

    def reanalyze_pair(self, doc1, doc2, sim: float, on_success, on_error):
        worker = Worker(reanalyze_pair, doc1, doc2, sim)
        worker.signals.result.connect(on_success)
        worker.signals.error.connect(on_error)
        pool = QThreadPool.globalInstance()
        pool.start(worker)


class Worker(QRunnable):
    def __init__(self, fn, *args, **kwargs):
        super(Worker, self).__init__()
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.signals = WorkerSignals()

    @Slot()
    def run(self):
        try:
            result = self.fn(*self.args, **self.kwargs)
        except Exception:
            exctype, value = sys.exc_info()[:2]
            self.signals.error.emit((exctype, value, traceback.format_exc()))
        else:
            self.signals.result.emit(result)


class WorkerSignals(QObject):
    error = Signal(tuple)
    result = Signal(object)



from __future__ import annotations

from networkx import Graph, articulation_points, is_empty

from plagdef.model.models import Cluster, RatedCluster


class ClusterFilter:
    def __init__(self, min_cluster_char_len: int):
        self._min_cluster_char_len = min_cluster_char_len

    def filter(self, clusters: set[Cluster]):
        resolved_clusters = _resolve_overlaps(clusters)
        return self._remove_small_clusters(resolved_clusters)

    def _remove_small_clusters(self, clusters: set[Cluster]) -> set[Cluster]:
        filtered_clusters = set()
        for cluster in clusters:
            cluster_char_lengths = cluster.char_lengths()
            if cluster_char_lengths[0] >= self._min_cluster_char_len \
                and cluster_char_lengths[1] >= self._min_cluster_char_len:
                filtered_clusters.add(cluster)
        return filtered_clusters


def _resolve_overlaps(clusters: set[Cluster]) -> set[Cluster]:
    overlap_graph = _build_overlap_graph(clusters)
    cluster = _next_overlapping_cluster(overlap_graph)
    while cluster:
        ol_clusters = overlap_graph.adj[cluster]
        best_rated_cluster = RatedCluster(cluster, 0, 0)
        for ol_cluster in ol_clusters:
            better_rated_cluster = cluster.best_with_respect_to(ol_cluster)
            if better_rated_cluster > best_rated_cluster:
                best_rated_cluster = better_rated_cluster
        if cluster == best_rated_cluster:
            overlap_graph.remove_nodes_from(ol_clusters)
        else:
            overlap_graph.remove_node(cluster)
        cluster = _next_overlapping_cluster(overlap_graph)
    return set(overlap_graph)


def _build_overlap_graph(clusters: set[Cluster]) -> Graph:
    graph = Graph()
    for cluster in clusters:
        graph.add_node(cluster)
        for ol_cluster in clusters:
            if cluster != ol_cluster and cluster.overlaps_with(ol_cluster):
                graph.add_edge(cluster, ol_cluster)
    return graph


def _next_overlapping_cluster(graph: Graph):
    cluster = None
    if not is_empty(graph):
        try:
            cluster = next(articulation_points(graph))
        except StopIteration:
            cluster = max(graph.degree, key=lambda x: x[1])[0]
    return cluster



from __future__ import annotations

from collections import Counter, defaultdict
from collections.abc import Iterable
from dataclasses import dataclass
from enum import Enum
from functools import total_ordering
from pathlib import Path

from sortedcontainers import SortedSet

from plagdef import util


class File:
    def __init__(self, path: Path, content: any, binary: bool):
        self.path = path
        self.content = content
        self.binary = binary

    def __eq__(self, other):
        if type(other) is type(self):
            return self.content == other.content
        return False

    def __hash__(self):
        return hash(self.content)

    def __repr__(self):
        return f"File('{self.path.stem}')"


class Document:
    def __init__(self, name: str, path: str, text: str):
        self.name = name
        self.path = path
        self.text = text
        self.lang = None
        self.vocab = Counter()  # <lemma, sent_freq>
        self.urls = set()
        self._sents = SortedSet()

    def add_sent(self, sent: Sentence):
        self._sents.add(sent)

    def remove_sent(self, sent: Sentence):
        self._sents.remove(sent)

    def sents(self, include_common=False):
        if include_common:
            return self._sents
        else:
            return filter(lambda sent: not sent.common, self._sents)

    def __eq__(self, other):
        if type(other) is type(self):
            return self.text == other.text
        return False

    def __hash__(self):
        return hash(self.text)

    def __repr__(self):
        return f"Document('{self.name}')"

    def __getstate__(self):
        return self.name, self.path, self.text, self.vocab, self.urls, tuple(self._sents)

    def __setstate__(self, state):
        self.name, self.path, self.text, self.vocab, self.urls, sents = state
        self._sents = SortedSet()
        for sent in sents:
            sent.doc = self
            self._sents.add(sent)


@total_ordering
class Fragment:
    def __init__(self, start_char: int, end_char: int, doc: Document):
        self.start_char = start_char
        self.end_char = end_char
        self.doc = doc
        self.text = doc.text[start_char:end_char]

    def overlaps_with(self, other: Fragment) -> bool:
        if self.doc == other.doc:
            own_interval = set(range(self.start_char, self.end_char))
            other_interval = set(range(other.start_char, other.end_char))
            return len(own_interval.intersection(other_interval)) > 0
        return False

    def __eq__(self, other):
        if type(other) is type(self):
            return self.start_char == other.start_char and self.end_char == other.end_char and self.doc == other.doc
        return False

    def __lt__(self, other) -> bool:
        return self.start_char < other.start_char

    def __hash__(self):
        return hash((self.start_char, self.doc))

    def __len__(self):
        return self.end_char - self.start_char

    def __repr__(self):
        return f'Fragment({self.doc.name}, {self.start_char}, {self.end_char})'


class Sentence(Fragment):
    def __init__(self, start_char: int, end_char: int, bow: Counter, doc: Document):
        super().__init__(start_char, end_char, doc)
        self.words = []
        self.bow = bow
        self.tf_isf_bow = {}
        self.common = False

    @property
    def idx(self):
        return self.doc.sents(include_common=True).index(self)

    def adjacent_to(self, other: Sentence, adjacent_sents_gap: int) -> bool:
        return abs(self.idx - other.idx) - 1 <= adjacent_sents_gap


class Word(Fragment):
    def __init__(self, start_char: int, end_char: int, sent: Sentence):
        super().__init__(start_char, end_char, sent.doc)
        self.sent = sent


@dataclass(frozen=True)
class Seed:
    sent1: Sentence
    sent2: Sentence
    cos_sim: float
    dice_sim: float

    def __repr__(self):
        return f'Seed({self.sent1.idx}, {self.sent2.idx}, {self.cos_sim}, {self.dice_sim})'


class Cluster:
    def __init__(self, seeds: set[Seed]):
        self.seeds = frozenset(seeds)
        self.doc1 = next(iter(self.seeds)).sent1.doc
        self.doc2 = next(iter(self.seeds)).sent2.doc
        self.sents_doc1 = tuple(self._sents(in_first_doc=True))
        self.sents_doc2 = tuple(self._sents(in_first_doc=False))
        self.tf_isf_bow_doc1 = self._tf_isf_bow(doc1_sents=True)
        self.tf_isf_bow_doc2 = self._tf_isf_bow(doc1_sents=False)
        self.cos_sim = util.cos_sim(self.tf_isf_bow_doc1, self.tf_isf_bow_doc2)

    def _sents(self, in_first_doc: bool) -> list[Sentence]:
        sent_idc = [seed.sent1.idx for seed in self.seeds] if in_first_doc else [seed.sent2.idx for seed in self.seeds]
        start = min(sent_idc)
        end = max(sent_idc)
        # A sent idx of a seed can never be a common sent
        return list(self.doc1.sents())[start:end + 1] if in_first_doc \
            else list(self.doc2.sents())[start:end + 1]

    def _tf_isf_bow(self, doc1_sents: bool) -> dict:
        sents = self.sents_doc1 if doc1_sents else self.sents_doc2
        sent_vec_sum = defaultdict(lambda: 0.0)
        for sent in sents:
            for lemma, tf_isf_val in sent.tf_isf_bow.items():
                sent_vec_sum[lemma] += tf_isf_val
        return sent_vec_sum

    def overlaps_with(self, other: Cluster) -> bool:
        """Contrary to Sanchez-Perez et al.'s algorithm,
        two clusters are considered overlapping if and only if
        they share at least one sentence in doc1 and doc2."""
        return any([sent in other.sents_doc1 for sent in self.sents_doc1]) \
               and any([sent in other.sents_doc2 for sent in self.sents_doc2])

    def best_with_respect_to(self, ol_cluster: Cluster) -> RatedCluster:
        """Pick the best of the two overlapping clusters depending on their quality and size."""
        own = self._best_variant(ol_cluster)
        ol_cluster = ol_cluster._best_variant(self)
        return own if own >= ol_cluster else ol_cluster

    def _best_variant(self, ol_cluster: Cluster) -> RatedCluster:
        """Pick the best variant of the two possible qualities of a cluster.
        a: For each of the clusters' overlapping sentences in doc1 the algorithm looks for the most similar
        sentence in doc2 (from which doc1 potentially plagiarized this sentence). If this quality a computed with the
        clusters' overlapping fragments in doc1 combined with this cluster's non-overlapping fragment in doc1 is higher
        than or equal to b it is returned.
        b: For each of the clusters' overlapping sentences in doc2 the algorithm looks for the most similar
        sentence in doc1 (from which doc2 potentially plagiarized this sentence). If this quality b computed with the
        clusters' overlapping fragments in doc2 combined with this cluster's non-overlapping fragment in doc2 is higher
        than a it is returned."""
        a = self._rate_with_respect_to(ol_cluster, first_doc_susp=True)
        b = self._rate_with_respect_to(ol_cluster, first_doc_susp=False)
        return a if a >= b else b

    def _rate_with_respect_to(self, ol_cluster: Cluster, first_doc_susp: bool) -> RatedCluster:
        """Compute the quality of this cluster in respect to an overlapping cluster ol_cluster according to the
        formula:
        q_{ol_cluster}(self) = sim_{self.sents_doc2}(O) + (1 - sim_{self.sents_doc2}(O))
        + sim_{self.sents_doc2}(N), O = self.sents_doc1 ∩ ol_cluster.sents_doc1 being the overlap in doc1
        and N = self.sents_doc1 / O the non-overlapping fragment in this cluster. The parameter first_doc_susp exchanges
        doc1 and doc2 in the formula."""
        own_sents_susp = self.sents_doc1 if first_doc_susp else self.sents_doc2
        own_sents_src = self.sents_doc2 if first_doc_susp else self.sents_doc1
        other_sents_susp = ol_cluster.sents_doc1 if first_doc_susp else ol_cluster.sents_doc2
        ol_sents_susp = set(filter(lambda susp_sent: susp_sent in other_sents_susp, own_sents_susp))
        own_non_ol_sents_susp = set(own_sents_susp).difference(ol_sents_susp)
        ol_sim = self._fragment_similarity(ol_sents_susp, own_sents_src)
        non_ol_sim = self._fragment_similarity(own_non_ol_sents_susp, own_sents_src)
        return RatedCluster(self, ol_sim + (1 - ol_sim) * non_ol_sim, len(own_sents_susp))

    def _fragment_similarity(self, fragment_sents: Iterable[Sentence], cluster_sents: Iterable[Sentence]) -> float:
        """Compute the non-symmetric similarity of a cluster's fragment in document a (fragment_sents) to the whole
        cluster fragment in document b (cluster_sents) according to the formula:
        sim_{cluster_sents}(fragment_sents) = 1/|fragment_sents|
            * sum_{s in fragment_sents}(max_{r in cluster_sents}(cos(s,r)))
        """
        similarity = 0.0
        frag_sents_len = 0
        for frag_sent in fragment_sents:
            similarity += max([util.cos_sim(frag_sent.tf_isf_bow, cluster_sent.tf_isf_bow)
                               for cluster_sent in cluster_sents])
            frag_sents_len += 1
        return similarity / frag_sents_len if frag_sents_len else 0

    def char_lengths(self) -> tuple[int, int]:
        char_len_doc1 = sum(len(sent) for sent in self.sents_doc1)
        char_len_doc2 = sum(len(sent) for sent in self.sents_doc2)
        return char_len_doc1, char_len_doc2

    def __eq__(self, other):
        if type(other) is type(self):
            return self.seeds == other.seeds
        return False

    def __hash__(self):
        return hash(self.seeds)

    def __repr__(self):
        return f'Cluster({len(self.seeds)}, {self.cos_sim})'


@total_ordering
class RatedCluster:
    def __init__(self, cluster: Cluster, quality: float, size: int):
        self.cluster = cluster
        self.quality = quality
        self.size = size  # number of sentences

    def __lt__(self, other: RatedCluster) -> bool:
        if (self.quality > 0.99 and other.quality > 0.99) or self.quality == other.quality:
            return self.size < other.size
        else:
            return self.quality < other.quality

    def __eq__(self, other):
        if type(other) is type(self):
            return self.quality == other.quality and self.size == other.size and self.cluster == other.cluster
        return False


class Match:
    def __init__(self, match_type: MatchType, frag1: Fragment, frag2: Fragment):
        self.type = match_type
        if frag1.doc == frag2.doc:
            raise SameDocumentError(f"Cannot match the document {frag1.doc} with itself. Please check whether your "
                                    f"archive and document files overlap.")
        self.frag_pair = frozenset({frag1, frag2})

    @classmethod
    def from_cluster(cls, match_type: MatchType, cluster: Cluster) -> Match:
        text_doc1_start, text_doc1_end = cluster.sents_doc1[0].start_char, cluster.sents_doc1[-1].end_char
        text_doc2_start, text_doc2_end = cluster.sents_doc2[0].start_char, cluster.sents_doc2[-1].end_char
        return Match(match_type, Fragment(text_doc1_start, text_doc1_end, cluster.doc1),
                     Fragment(text_doc2_start, text_doc2_end, cluster.doc2))

    def overlaps_with(self, other: Match) -> bool:
        frag1, frag2 = self.frag_pair
        return any([frag1.overlaps_with(other_frag) for other_frag in other.frag_pair]) \
               and any([frag2.overlaps_with(other_frag) for other_frag in other.frag_pair])

    def frag_from_doc(self, doc: Document):
        return next(filter(lambda frag: frag.doc == doc, self.frag_pair), None)

    def __eq__(self, other):
        if type(other) is type(self):
            return self.frag_pair == other.frag_pair
        return False

    def __hash__(self):
        return hash(self.frag_pair)

    def __len__(self):
        return sum(len(frag) for frag in self.frag_pair)

    def __repr__(self):
        if len(self.frag_pair) < 2:
            return 'Match()'
        frag1, frag2 = self.frag_pair
        return f'Match({frag1}, {frag2})'


class DocumentPairMatches:
    def __init__(self, doc1: Document, doc2: Document, matches: Iterable = None):
        self.doc1 = doc1
        self.doc2 = doc2
        self._matches = defaultdict(set)
        if matches:
            self.update(matches)

    @classmethod
    def from_matches(cls, matches: [Match]) -> [DocumentPairMatches]:
        doc_pair_matches = []
        dpm_dict = defaultdict(list)
        for match in matches:
            frag1, frag2 = match.frag_pair
            dpm_dict[frozenset({frag1.doc, frag2.doc})].append(match)
        for pair in dpm_dict.keys():
            doc1, doc2 = pair
            doc_pair_matches.append(DocumentPairMatches(doc1, doc2, dpm_dict[pair]))
        return doc_pair_matches

    def update(self, matches: Iterable):
        [self.add(match) for match in matches]

    def add(self, match: Match):
        frag1, frag2 = match.frag_pair
        if not {self.doc1, self.doc2} == {frag1.doc, frag2.doc}:
            raise DifferentDocumentPairError(f'Only matches of document pair ({self.doc1}, {self.doc2})'
                                             f'can be added.')
        self._matches[str(match.type)].add(match)

    def list(self, match_type: MatchType) -> set[Match]:
        if str(match_type) in self._matches:
            return self._matches[str(match_type)]
        return set()

    def __len__(self):
        return sum(len(m) for m in self._matches.values())

    def __repr__(self):
        return f'DocPairMatches({self.doc1}, {self.doc2}, {len(self)})'

    def __eq__(self, other):
        if type(other) is type(self):
            return {self.doc1, self.doc2} == {other.doc1, other.doc2}

    def __hash__(self):
        return hash(frozenset({self.doc1, self.doc2}))


class MatchType(Enum):
    VERBATIM = 0
    INTELLIGENT = 1
    SUMMARY = 2

    def __str__(self):
        return self.name.lower()


class DifferentDocumentPairError(Exception):
    pass


class SameDocumentError(Exception):
    pass



from __future__ import annotations

from dataclasses import dataclass

from plagdef.model.models import Document, DocumentPairMatches, Match, MatchType, Cluster, Fragment
from plagdef.model.pipeline.extension import ClusterBuilder
from plagdef.model.pipeline.filtering import ClusterFilter
from plagdef.model.pipeline.seeding import SeedFinder


@dataclass(frozen=True)
class PipeComponents:
    seeder: SeedFinder
    verbatim_matcher: VerbatimMatcher
    intelligent_cb: ClusterBuilder
    summary_cb: ClusterBuilder
    cf: ClusterFilter


class Pipeline:
    def __init__(self, doc1: Document, doc2: Document, pipe_comps: PipeComponents):
        self._doc1 = doc1
        self._doc2 = doc2
        self._pipe_comps = pipe_comps

    def find_matches(self) -> DocumentPairMatches:
        doc_pair_matches = DocumentPairMatches(self._doc1, self._doc2)
        seeds = self._pipe_comps.seeder.seed(self._doc1, self._doc2)
        clusters = self._build_clusters(seeds, self._pipe_comps.intelligent_cb)
        verbatim_matches = intelligent_matches = summary_matches = set()
        if len(clusters):
            verbatim_matches = self._pipe_comps.verbatim_matcher.find_verbatim_matches(clusters)
            intelligent_matches = {Match.from_cluster(MatchType.INTELLIGENT, cluster) for cluster in clusters}
        summary_clusters = self._build_clusters(seeds, self._pipe_comps.summary_cb)
        if len(summary_clusters):
            sum_cluster_len_doc1, sum_cluster_len_doc2 = \
                tuple(map(sum, zip(*(cluster.char_lengths() for cluster in summary_clusters))))
            if sum_cluster_len_doc1 >= 3 * sum_cluster_len_doc2 \
                or sum_cluster_len_doc2 >= 3 * sum_cluster_len_doc1:
                summary_matches = {Match.from_cluster(MatchType.SUMMARY, cluster) for cluster in summary_clusters}
        doc_pair_matches.update({*verbatim_matches, *intelligent_matches, *summary_matches})
        return doc_pair_matches

    def _build_clusters(self, seeds, cluster_builder: ClusterBuilder):
        clusters = cluster_builder.extend(seeds)
        clusters = self._pipe_comps.cf.filter(clusters)
        return clusters


class VerbatimMatcher:
    def __init__(self, min_verbatim_match_char_len: int):
        self._min_verbatim_match_char_len = min_verbatim_match_char_len

    def find_verbatim_matches(self, clusters: set[Cluster]) -> set[Match]:
        matches = set()
        for cluster in clusters:
            cluster_matches = self._common_words(cluster)
            matches.update(VerbatimMatcher._resolve_match_overlaps(cluster_matches))
        return matches

    def _common_words(self, cluster: Cluster) -> set[Match]:
        verbatim_matches = set()
        frag1_words = [word for sent_words in [sent.words for sent in cluster.sents_doc1] for word in sent_words]
        frag2_words = [word for sent_words in [sent.words for sent in cluster.sents_doc2] for word in sent_words]
        lookup = [[0 for _ in range(len(frag2_words) + 1)] for _ in range(len(frag1_words) + 1)]
        for i in range(1, len(frag1_words) + 1):
            for j in range(1, len(frag2_words) + 1):
                if frag1_words[i - 1].text.lower() == frag2_words[j - 1].text.lower():
                    lookup[i][j] = lookup[i - 1][j - 1] + 1
                    match_char_len = sum([len(word) for word in frag1_words[i - lookup[i][j]:i]])
                    if match_char_len >= self._min_verbatim_match_char_len:
                        frag1_first, frag2_first = i - lookup[i][j], j - lookup[i][j]
                        # Include punctuation (mostly periods) if exists
                        punct = VerbatimMatcher._include_punct(cluster, frag1_words[i - 1].end_char,
                                                               frag2_words[j - 1].end_char)
                        frag1 = Fragment(frag1_words[frag1_first].start_char, frag1_words[i - 1].end_char + punct,
                                         cluster.doc1)
                        frag2 = Fragment(frag2_words[frag2_first].start_char, frag2_words[j - 1].end_char + punct,
                                         cluster.doc2)
                        verbatim_matches.add(Match(MatchType.VERBATIM, frag1, frag2))
        return verbatim_matches

    @classmethod
    def _resolve_match_overlaps(cls, matches: set[Match]) -> set[Match]:
        non_ol_matches = set()
        for match in sorted(matches, key=len, reverse=True):
            if not any(match.overlaps_with(non_ol_match) for non_ol_match in non_ol_matches):
                non_ol_matches.add(match)
        return non_ol_matches

    @classmethod
    def _include_punct(cls, cluster: Cluster, frag1_end_char: int, frag2_end_char: int) -> int:
        text_ends_with_word = frag1_end_char == len(cluster.doc1.text) or frag2_end_char == len(cluster.doc2.text)
        if not text_ends_with_word and cluster.doc1.text[frag1_end_char] == cluster.doc2.text[frag2_end_char] \
            and not cluster.doc1.text[frag1_end_char].isspace():
            return 1
        return 0



from __future__ import annotations

import os
import string
from collections import Counter
from functools import partial
from urllib.parse import urlparse

import stanza
from stanza import Pipeline
from tqdm.contrib.concurrent import thread_map
from urlextract import URLExtract

from plagdef.model import stopwords
from plagdef.model.models import Document, Sentence, Word

PRCS = 'tokenize,mwt,pos,lemma'
PIPE_LVL = 'WARN'
LOAD_LVL = 'INFO'

# Update TLDs
URLExtract().update_when_older(7)


class Preprocessor:
    def __init__(self, min_sent_len: int, rem_stop_words: bool):
        self._min_sent_len = min_sent_len
        self._rem_stop_words = rem_stop_words

    def preprocess(self, lang: str, docs: set[Document], common_docs: list[Document] = None):
        nlp_model = _nlp_pipe(lang)
        stop_words = stopwords.ENGLISH if lang == 'en' else stopwords.GERMAN
        common_word_lists = _common_word_lists(nlp_model, common_docs) if common_docs else []
        thread_map(partial(self._preprocess, nlp_model=nlp_model, common_word_lists=common_word_lists,
                           stop_words=stop_words), docs, max_workers=os.cpu_count(),
                   total=len(docs), desc='Preprocessing', unit='doc')

    def _preprocess(self, doc: Document, nlp_model: Pipeline, common_word_lists: list[list[str]], stop_words: set[str]):
        sents = nlp_model(doc.text).sentences
        for sent_idx, sent in enumerate(sents):
            filtered_words = _word_filter(sent.words)
            if self._rem_stop_words:
                sent_lemmas = [word.lemma for word in filtered_words if word.text.lower() not in stop_words]
            else:
                sent_lemmas = [word.lemma for word in filtered_words]
            if len(sent_lemmas):
                lemma_count = Counter(sent_lemmas)
                sentence = Sentence(sent.tokens[0].start_char, sent.tokens[-1].end_char, lemma_count, doc)
                sentence.words = [Word(word.parent.start_char, word.parent.end_char, sentence)
                                  for word in filtered_words]
                doc.add_sent(sentence)
                if _sent_contains_common_words(sentence.words, common_word_lists):
                    sentence.common = True
                else:
                    for lemma in lemma_count.keys():
                        doc.vocab[lemma] += 1
        self._join_small_sentences(doc)
        self._remove_small_sentences(doc)
        _extract_urls(doc)

    def _join_small_sentences(self, doc: Document):
        sents = doc.sents(include_common=True)
        idx, sent_count = 0, len(sents)
        while idx < sent_count - 1:
            sent1, sent2 = sents[idx], sents[idx + 1]
            if (not sent1.common and not sent2.common) and \
                (len(sent1.words) < self._min_sent_len or
                 (sent2 == sents[-1] and len(sent2.words) < self._min_sent_len)):
                for lemma in sent1.bow.keys():
                    if lemma in sent2.bow:
                        doc.vocab[lemma] -= 1
                joined_sent = Sentence(sent1.start_char, sent2.end_char, sent1.bow + sent2.bow, doc)
                joined_sent.words = sent1.words + sent2.words
                doc.remove_sent(sent1), doc.remove_sent(sent2)
                doc.add_sent(joined_sent)
                idx -= 1
                sent_count -= 1
            idx += 1

    def _remove_small_sentences(self, doc: Document):
        sents = list(doc.sents())
        idx, sent_count = 0, len(sents)
        while idx < sent_count:
            if len(sents[idx].words) < self._min_sent_len:
                for lemma in sents[idx].bow.keys():
                    doc.vocab[lemma] -= 1
                doc.remove_sent(sents[idx]), sents.remove(sents[idx])
                idx -= 1
                sent_count -= 1
            idx += 1
        doc.vocab += Counter()  # Remove zero counts


def _extract_urls(doc: Document, extractor=URLExtract()):
    urls = extractor.find_urls(doc.text, only_unique=True, check_dns=True)
    for url in urls:
        url = url[:-1] if url[-1] in string.punctuation else url
        parsed_url = urlparse(url, "https")
        if parsed_url.scheme in ("http", "https"):
            doc.urls.add(parsed_url.geturl().rstrip("/").replace("///", "//"))


def _nlp_pipe(lang: str) -> Pipeline:
    try:
        return stanza.Pipeline(lang, processors=PRCS, logging_level=PIPE_LVL)
    except:  # Unpickling error raises Exception, cannot narrow
        stanza.download(lang, processors=PRCS, logging_level=LOAD_LVL)
        return stanza.Pipeline(lang, processors=PRCS, logging_level=PIPE_LVL)


def _common_word_lists(pipe: Pipeline, common_docs: list[Document]) -> list[list[str]]:
    common_word_lists = []
    for doc in common_docs:
        for line in doc.text.splitlines():
            parsed_line = pipe(line)
            line_words = []
            for sent in parsed_line.sentences:
                [line_words.append(word.text.lower()) for word in _word_filter(sent.words)]
            common_word_lists.append(line_words) if len(line_words) else None
    return common_word_lists


def _word_filter(stanza_words: list) -> list:
    return [word for word in stanza_words if not word.upos == 'PUNCT' and word.text.isalnum() and len(word.text) > 1]


def _sent_contains_common_words(sent_words: list[Word], common_word_lists: list[list[str]]) -> bool:
    sent_word_texts = [word.text.lower() for word in sent_words]
    for common_word_list in common_word_lists:
        if all(common_word in sent_word_texts for common_word in common_word_list):
            return True
    return False



from __future__ import annotations

import math
from typing import Union

from plagdef import util
from plagdef.model.models import Seed
from plagdef.model.pipeline.preprocessing import Sentence, Document


class SeedFinder:
    def __init__(self, min_cos_sim: float, min_dice_sim: float):
        self._min_cos_sim = min_cos_sim
        self._min_dice_sim = min_dice_sim

    def seed(self, doc1: Document, doc2: Document) -> set[Seed]:
        _vectorize_sents(doc1, doc2)
        seeds = set()
        for doc1_sent in doc1.sents():
            for doc2_sent in doc2.sents():
                seed = self._match(doc1_sent, doc2_sent)
                if seed:
                    seeds.add(seed)
        return seeds

    def _match(self, sent1: Sentence, sent2: Sentence) -> Union[Seed, None]:
        cos_sim = util.cos_sim(sent1.tf_isf_bow, sent2.tf_isf_bow)
        dice_sim = util.dice_sim(sent1.tf_isf_bow, sent2.tf_isf_bow)
        if cos_sim > self._min_cos_sim and dice_sim > self._min_dice_sim:
            return Seed(sent1, sent2, cos_sim, dice_sim)


def _vectorize_sents(doc1: Document, doc2: Document):
    """
    Compute the tf-isf = tf x ln(N/sf), N being the number of sentences in corpus
    and sf the number of sentences containing the term
    """
    sf = doc1.vocab + doc2.vocab
    doc1_sents, doc2_sents = list(doc1.sents()), list(doc2.sents())
    N = len(doc1_sents) + len(doc2_sents)
    for sent in doc1_sents:
        for lemma in sent.bow:
            sent.tf_isf_bow[lemma] = sent.bow[lemma] * math.log(N / float(sf[lemma]))
    for sent in doc2_sents:
        for lemma in sent.bow:
            sent.tf_isf_bow[lemma] = sent.bow[lemma] * math.log(N / float(sf[lemma]))



from __future__ import annotations

from plagdef.model.models import DocumentPairMatches, MatchType


def generate_text_report(matches: list[DocumentPairMatches]) -> str:
    report = ''
    for match_type in MatchType:
        wrote_heading = False
        for doc_pair_matches in matches:
            typed_matches = doc_pair_matches.list(match_type)
            if len(typed_matches):
                if not wrote_heading:
                    report += f'{str(match_type).capitalize()} matches:\n'
                    wrote_heading = True
                report += f"  Pair('{doc_pair_matches.doc1.path}', '{doc_pair_matches.doc2.path}'):\n"
                for match in sorted(typed_matches, key=lambda m: m.frag_from_doc(doc_pair_matches.doc1).start_char):
                    frag1, frag2 = match.frag_from_doc(doc_pair_matches.doc1), match.frag_from_doc(
                        doc_pair_matches.doc2)
                    report += f'    Match(Fragment({frag1.start_char}, {frag1.end_char}), Fragment(' \
                              f'{frag2.start_char}, {frag2.end_char}))\n'
    intro = 'No matches found.'
    if report:
        intro = f'Found {len(matches) if len(matches) else "no"} suspicious document pair' \
                f'{"s" if len(matches) > 1 else ""}.\n' \
                'Reporting matches for each pair like this:\n' \
                f'  Match(Fragment(start_char, end_char), Fragment(start_char, end_char))\n\n'
    return intro + report



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['plagdef', 'plagdef.gui', 'plagdef.model', 'plagdef.model.pipeline']

package_data = \
{'': ['*'], 'plagdef': ['config/*'], 'plagdef.gui': ['res/*', 'ui/*']}

install_requires = \
['Werkzeug>=2.0,<3.0',
 'beautifulsoup4>=4.10,<5.0',
 'click>=8.0,<9.0',
 'deep-translator>=1.7,<2.0',
 'jsonpickle>=2.1,<3.0',
 'langdetect>=1.0,<2.0',
 'networkx>=2.7,<3.0',
 'numpy>=1.22,<2.0',
 'ocrmypdf>=13.4,<14.0',
 'pdfplumber>=0.6,<0.7',
 'pyside6',
 'python-magic>=0.4,<0.5',
 'sortedcontainers>=2.4,<3.0',
 'stanza>=1.3,<2.0',
 'tqdm>=4.63,<5.0',
 'urlextract>=1.5,<2.0']

entry_points = \
{'console_scripts': ['plagdef = plagdef.app:cli',
                     'plagdef-gui = plagdef.app:gui']}

setup_kwargs = {
    'name': 'plagdef',
    'version': '1.4.10',
    'description': 'A tool which makes life hard for students who try to make theirs simple.',
    'long_description': "# PlagDef\n\n[![PyPI version](https://badge.fury.io/py/plagdef.svg)](https://badge.fury.io/py/plagdef)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/plagdef)\n![GitHub](https://img.shields.io/github/license/devWhyqueue/plagdef)\n[![Test](https://github.com/devWhyqueue/plagdef/actions/workflows/cd.yml/badge.svg)](https://github.com/devWhyqueue/plagdef/actions/workflows/test.yml)\n[![Coverage](https://sonarcloud.io/api/project_badges/measure?project=devWhyqueue_plagdef&metric=coverage)](https://sonarcloud.io/dashboard?id=devWhyqueue_plagdef)\n\nA tool which makes life hard for students who try to make theirs simple.\n\n# Installation\n\nGet it from PyPI:\n\n```\n$ pip install plagdef\n````\n\nOr build it yourself:\n\n```\n$ git clone git://github.com/devWhyqueue/plagdef\n$ poetry build\n$ pip install dist/plagdef-{VERSION}.tar.gz\n````\n\n# Requirements\n\n## OCRMyPDF\n\nThis library is used for improved PDF text extraction.\\\nTo install its necessary dependencies for your operating system take a look at:\\\nhttps://ocrmypdf.readthedocs.io/en/latest/installation.html\n\nAnd don't forget to download the German language pack to your _tessdata_ folder from here:\\\nhttps://github.com/tesseract-ocr/tessdata\n\n## Libmagic\n\n**After** (important!) you installed PlagDef, install the libmagic library.\\\nPlagDef uses it to detect character encodings.\\\nFurther instructions can be found here:\\\nhttps://github.com/ahupp/python-magic#installation\n\n# Usage\n\nRun the GUI:\n\n```\n$ plagdef-gui\n````\n\nOr if you prefer a CLI:\n\n```\n$ plagdef -h\n````\n\n# Development\n\nClone the repo and install dependencies:\n\n```\n$ git clone git://github.com/devWhyqueue/plagdef\n$ poetry install\n````\n\n# Publish to PyPI\n\nIn your virtual environment build and upload PlagDef:\n\n```\n$ poetry publish --build\n````\n",
    'author': 'Yannik Queisler',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/devWhyqueue/plagdef',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<3.11',
}


setup(**setup_kwargs)



from __future__ import annotations

import logging
from itertools import combinations, product

from tqdm import tqdm

from plagdef.model import matching
from plagdef.model.matching import PipeComponents, VerbatimMatcher
from plagdef.model.models import Document, DocumentPairMatches
from plagdef.model.pipeline.extension import ClusterBuilder
from plagdef.model.pipeline.filtering import ClusterFilter
from plagdef.model.pipeline.preprocessing import Preprocessor
from plagdef.model.pipeline.seeding import SeedFinder
from plagdef.util import parallelize

log = logging.getLogger(__name__)


class DocumentMatcher:
    def __init__(self, config: dict):
        self._preprocessor = Preprocessor(config['min_sent_len'], config['rem_stop_words'])
        self._seeder = SeedFinder(config['min_cos_sim'], config['min_dice_sim'])
        self._verbatim_matcher = VerbatimMatcher(config['min_verbatim_match_char_len'])
        self._intelligent_cb = ClusterBuilder(config['adjacent_sents_gap'], config['min_adjacent_sents_gap'],
                                              config['min_sent_number'], config['min_cluster_cos_sim'])
        self._summary_cb = ClusterBuilder(config['adjacent_sents_gap_summary'], config['min_adjacent_sents_gap'],
                                          config['min_sent_number'], config['min_cluster_cos_sim'])
        self._cluster_filter = ClusterFilter(config['min_cluster_char_len'])

    def preprocess(self, lang: str, docs: set[Document], common_docs=None):
        self._preprocessor.preprocess(lang, docs, common_docs)

    def find_matches(self, docs: set[Document], archive_docs=None) -> list[DocumentPairMatches]:
        doc_combs = set(combinations(docs, 2))
        if archive_docs:
            doc_overlap = docs.intersection(archive_docs)
            log.warning(f'The following documents have counterparts with identical contents in the archive: '
                        f'[{str(doc_overlap)[1:-1]}]') if len(doc_overlap) else None
            doc_combs.update(product(docs, archive_docs.difference(doc_overlap)))
        return parallelize(self._find_matches, list(doc_combs))

    def _find_matches(self, doc_combs, pos=0) -> list[DocumentPairMatches]:
        matches = []
        for doc1, doc2 in tqdm(doc_combs, desc='Matching', unit='pair', total=len(doc_combs), position=pos,
                               leave=False):
            pipe = matching.Pipeline(doc1, doc2,
                                     PipeComponents(self._seeder, self._verbatim_matcher, self._intelligent_cb,
                                                    self._summary_cb, self._cluster_filter))
            doc_pair_matches = pipe.find_matches()
            matches.append(doc_pair_matches) if len(doc_pair_matches) else None
        return matches



from __future__ import annotations

import logging
from pathlib import Path

from click import UsageError

from plagdef.config import settings
from plagdef.model.detection import DocumentMatcher
from plagdef.model.download import download_all_external_sources
from plagdef.model.models import DocumentPairMatches, Document
from plagdef.model.pipeline.translate import translate, detect_lang, docs_in_other_langs
from plagdef.repositories import UnsupportedFileFormatError, DocumentPickleRepository, DocumentFileRepository, \
    FileRepository

log = logging.getLogger(__name__)


def find_matches(doc_repo, archive_repo=None, common_doc_repo=None, config=settings, download=True) \
    -> list[DocumentPairMatches]:
    try:
        doc_matcher = DocumentMatcher(config)
        archive_docs = None
        if archive_repo:
            archive_docs = _preprocess_docs(doc_matcher, config['ser'], archive_repo, common_doc_repo)
        docs = _preprocess_docs(doc_matcher, config['ser'], doc_repo, common_doc_repo)
        if download and config['download_path']:
            _save_all_external_sources(docs, config['download_path'])
            ext_docs = _preprocess_docs(doc_matcher,
                                        config['ser'], DocumentFileRepository(Path(config['download_path'])),
                                        common_doc_repo, trans=True)
            archive_docs = archive_docs.union(ext_docs) if archive_docs else ext_docs
        doc_pair_matches = doc_matcher.find_matches(docs, archive_docs)
        return doc_pair_matches
    except UnsupportedFileFormatError as e:
        raise UsageError(str(e)) from e


def _preprocess_docs(doc_matcher, use_serialization, doc_repo, common_doc_repo=None, trans=False) -> set[Document]:
    common_dir_path = common_doc_repo.base_path if common_doc_repo else None
    common_docs = common_doc_repo.list() if common_doc_repo else None
    docs = _translate_docs(doc_repo) if trans else doc_repo.list()
    if use_serialization:
        doc_ser = DocumentPickleRepository(doc_repo.base_path, common_dir_path)
        prep_docs = {d for d in doc_ser.list() if d in docs}
        unprep_docs = docs.difference(prep_docs)
        doc_matcher.preprocess(doc_repo.lang, unprep_docs, common_docs)
        preprocessed_docs = prep_docs.union(unprep_docs)
        doc_ser.save(preprocessed_docs)
    else:
        doc_matcher.preprocess(doc_repo.lang, docs, common_docs)
        preprocessed_docs = docs
    return preprocessed_docs


def _save_all_external_sources(docs, download_path):
    external_sources = download_all_external_sources(docs, Path(download_path))
    external_sources_repo = FileRepository(Path(download_path))
    external_sources_repo.save_all(external_sources)


def _translate_docs(doc_repo: DocumentFileRepository) -> set[Document]:
    docs = doc_repo.list()
    detect_lang(docs)
    translated_docs = translate(docs_in_other_langs(docs, doc_repo.lang), doc_repo.lang)
    if len(translated_docs):
        for doc in translated_docs:
            new_path = str(Path(doc.path).with_name(f"{doc.name}_trans.txt"))
            doc_repo.remove_all({doc})
            doc.path = new_path
        doc_repo.save_all(translated_docs)
    return {doc for doc in docs if doc.lang == doc_repo.lang}


def write_json_reports(matches: list[DocumentPairMatches], repo):
    [repo.save(m) for m in matches]



import logging.config
import warnings
from ast import literal_eval
from configparser import ConfigParser
from pathlib import Path

import pkg_resources


def _read_config(path) -> dict:
    parser = ConfigParser()
    parser.read(path)
    config = {}
    for section in parser.sections():
        typed_config = [(key, literal_eval(val)) for key, val in parser.items(section)]
        config.update(dict(typed_config))
    return config


# Logging config
logging_config = pkg_resources.resource_filename(__name__, str(Path('config/logging.ini')))
logging.config.fileConfig(logging_config, disable_existing_loggers=False)

# App config
app_config_path = pkg_resources.resource_filename(__name__, str(Path('config/app.ini')))
settings = _read_config(app_config_path)

# Ignore warnings from Torch
warnings.filterwarnings("ignore")
